<?php if (!defined('PmWiki')) exit();
/* 
   publish page collection as book script, version 2.1.2
   works with PmWiki 2.2.56 or above
   requires php 5.3 or above and is compatible with php 5.5

   Copyright 2003-2014 John Rankin john.rankin@affinity.co.nz
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This file turns a trail page into a typeset book of the contents, in a
   form suitable for printing, via the (:typeset-book:) directive.
*/

## epub directive, provides (:epub-<kind>[ "<option text>"]:)
Markup('epub' , 'directives', '/\(:epub-?(book|article)?(?:\s*"([^"]+)")?\s*:\)/',
       function ($m) use (&$pagename) {
         return ServeWikibookEpub($pagename,
                                  ($m[1] ? $m[1] : 'book'),
                                  ($m[2] ? $m[2] : 'EPUB'));
       } );

function ServeWikibookEpub($pagename, $kind, $opttxt) {
  global $UploadDir, $UploadUrlFmt, $UrlLinkEpubFmt, $FmtV, $TimeFmt,
         $UploadPrefixFmt, $ScriptUrl, $UrlLinkFmt, $format;
  if ($format=='pdf') return '';
  $txt = FmtPageName('$['.$opttxt.']', $pagename);
  $PDFKind = array('book'    => 'book',
                   'report'  => 'typeset',
                   'article' => 'include');
  $bn = $pagename.".epub";
  $epubpath = FmtPageName("$UploadDir$UploadPrefixFmt",    $pagename)."/$bn";
  $epuburl  = FmtPageName("$UploadUrlFmt$UploadPrefixFmt", $pagename)."/$bn";
  if (file_exists($epubpath)) {
    $stat = stat($epubpath);
    $FmtV['$LinkUrl'] = PUE($epuburl);
    $FmtV['$LinkText'] = FmtPageName('$Titlespaced', $pagename);
    $FmtV['$LinkAlt'] = FmtPagename('$[download EPUB]', $pagename);
    SDV($UrlLinkEpubFmt, "<span class='epub'>$UrlLinkFmt</span>");
    $lnk = '&emsp;'.FmtPageName($UrlLinkEpubFmt, $pagename).' ('.
      strftime($TimeFmt, $stat['mtime']).')';
  }
  $out[] = FmtPageName("<form class='publish' action='\$ScriptUrl' method='get'>", 
    $pagename);
  $outOptions = array('n'       => FmtPageName('$FullName',$pagename),
                      'format'  => 'epub',
                      'action'  => 'publish',
                      'ptype'   => $PDFKind[$kind],
                      'pdfkind' => $kind);
  foreach($outOptions as $name => $value)
    $out[] = '<input type="hidden" name="'.$name.'" value="'.$value.'" />';
  $out[] = "<input class='pubbutton' type='submit' value='$txt' />$lnk</form>";
  return Keep(implode('', $out));
}


# image helper function
if ($format=='epub') Markup('^img', 'block',
  "/^((?>(\\s+|%%|%[A-Za-z][-,=:#\\w\\s'\".]*%)*)$KeepToken(\\d+L)$KeepToken)(\\s*\\|\\s?)?(.*)$/",
    function ($m) {
      return (strpos($GLOBALS['KPV'][$m[3]],'<img')===false) ? $m[0] :
         '<:block,1>'.($m[4] ? '<figure>'.$m[1].'<br />'. 
                               '<figcaption>'.$m[5].'</figcaption></figure>'
                             : '<div>'.$m[1].$m[5].'</div>');
    } );

function HandleBookPublish($pagename, $auth = 'read') {
  global $HandlePublishFmt,$SpaceWikiWords,$HeadingDepth,$format;
  global $PublishFormPageFmt,$HTMLHeaderFmt,$PdfFrontMatterPattern,$TbookCustomMetadata;
  global $ExtractTrailStopsFunction,$PrintThisPageFunction,$XMLEncodingFmt;
  global $BookTrail,$PageAnchors;  $BookTrail = true;
  SDV($ExtractTrailStopsFunction, 'ExtractTrailStops');
  SDV($PrintThisPageFunction, 'PrintThisPage');
  $Tp    = 'part';  $Ta = 'appendix';
  $Tu    = $Tp.' class="unnumber"';
  $Tbook = array('chapter',
     'section','subsection','subsubsection','paragraph','subparagraph');
  $Chapter  = 'E:blockquote:';
  $TbookPTV = array('epigraph'=>'E', 'dedication'=>'K', 'about'=>'K', 'isbn'=>'I', 
                    'graphics'=>'V', 'year'=>'I');
  $TbookDir = array('description' => 'publisher', 'keywords' => 'keywords');
  $TbookCCL = array('rights' => array('^(none|cc|gfdl)' => 'CopyrightClassFmt',
                                      '^(cc\s+by(-nc)?(-(?:sa|nd))?(\s+[^"]+)?)' => 'CopyrightStyleFmt',
                                      '^(?:none|gfdl):?\s+([^"]+)?' => 'CopyrightStyleFmt',
                                      '"([^"]+)"$' => 'CopyrightCaptionFmt'),
                    'style'  => array('^(Attach:[^\s"\']+\.sty)$' => 'BookStyleFmt'));
  $TbookOpt = array('typeset-book' => 'SetPdfMetaTags',
                    'cfloat'       => 'SetCfloatName',
                    'blurb'        => 'SetBookBlurb');
  $TitleFmt = ($SpaceWikiWords) ? '$Titlespaced' : '$Title';
  $Unnumber = array('preface',  'introduction', 'acknowledgments', 'colophon', 'part',
                    'foreword', 'afterword',    'prologue',        'epilogue', 'unnumber');

  $trailname = $pagename;
  $trailpage = RetrieveAuthPage($pagename, $auth, true, READPAGE_CURRENT);
  if (!$trailpage) Abort('?cannot read $pagename');
  PCache($pagename,$trailpage);
  $trailpage['text'] = str_replace("\\\\\n",'[[&lt;&lt;]]',$trailpage['text']);
  if ($format=='pdf') {
      $Chapter = 'E:aphorism:';

## gobble chapter contributors
      Markup('contrib', 'inline', '/^Contributor:.*?$/', '');

## epigraph
      Markup('tbepi','block','/^E:([^:]+):(.*?)(?:\s*&mdash;\s*(.*?)\s*("[^"]+")?)?$/',
  function ($m) {
    return '<:block><tbook:'.$m[1].'>'.$m[2].
  ($m[3] ? '<tbook:caption'.($m[4] ? ' caption='.$m[4] : '').'>'.$m[3].'</tbook:caption>' 
         : '').'</tbook:'.$m[1].'>';
  } );

## frontispiece
      Markup('tblogo','block','/^V:([^:]+):([^\s]+)\s*(.*?)$/',
  function ($m) {
    return '<:block>'.($m[3] ? SetFigureCaption('%id=wikipublisher.frontispiece%',$m[2],$m[3]).'</tbook:figure>' 
                             : $m[2]);
  } );

## dedication
      Markup('tbdedn','<^*','/^K:([^:]+):(.*?)$/',
  function ($m) {
    return '<:block><tbook:'.$m[1].'>'.str_replace(':+','<tbook:newline vspace="1ex"/>',$m[2]).'</tbook:'.$m[1].'>';
  } );

## publication year
      Markup('tbyear','block','/^I:([^:]+):(.*?)\s*("[^"]+")?$/',
  function ($m) {
    return '<:block><tbook:'.$m[1].($m[3] ? ' caption='.$m[3] : '').'>'.$m[2].'</tbook:'.$m[1].'>';
  } );

## extensible metadata
      Markup('tbdata','block','/^J:([^:]+):([^:]+):(.*?)$/',
          '<:block><tbook:$1 id="$2">$3</tbook:$1>');

      preg_match_all('/\\(:(\\w[-\\w]+):?(?!\\))(.*?):\\)/',$trailpage['text'],$dir,PREG_SET_ORDER);
      foreach ($dir as $k) $meta[$k[1]] = trim($k[2]);
      foreach ((array)$TbookCustomMetadata as $k) array_unshift($HTMLHeaderFmt, 
          "markup:J:titledata:$k:".(isset($meta[$k]) ? $meta[$k] : ''));
      foreach ($TbookPTV as $k => $v)
          if (isset($meta[$k])) array_unshift($HTMLHeaderFmt, "function:PrintMetaMarkup $v:$k:$meta[$k]");
      foreach ($TbookDir as $k => $v)
          if (isset($meta[$k])) array_unshift($HTMLHeaderFmt,
#             "\n<$v>".preg_replace('/&amp;([^;\s]+;)/',"&\\1",
#             htmlentities($meta[$k],ENT_NOQUOTES,$XMLEncodingFmt))."</$v>");
             "\n<$v>".PageVar($trailname, '$'.ucfirst($k))."</$v>");
      foreach ($TbookCCL as $k => $va)
          if (isset($meta[$k]))
              foreach ($va as $mat => $glo)
                  if (preg_match("/$mat/i",$meta[$k],$gv)) $GLOBALS[$glo] = 
                     ($k=='style') ? MakeLink($trailname, $gv[1], $gv[1], '', '$LinkUrl') : $gv[1];
      foreach ($TbookOpt as $k => $v) if (isset($meta[$k])) $v($meta[$k]);
      SDVA($HTMLHeaderFmt,
          PublishFormDefaults($trailname,$PublishFormPageFmt['book']));
  } else {
      if (preg_match("/\\(:typeset-book\\s*(.*?):\\)/",$trailpage['text'],$pdf))
          SetPdfMetaTags($pdf[1], false, false);
      Markup('tbchap','block','/^E:([^:]+):(.*?)(?:\s*&mdash;\s*(.*?)\s*(?:"([^"]+)")?)?$/',
  function ($m) {
    return '<:block><'.$m[1].'>'.$m[2].
        ($m[3] ? ' &mdash; <cite>'.$m[3].'</cite>' : '').
        ($m[4] ? ', <strong>'.$m[4].'</strong>'    : '').'</'.$m[1].'>';
  } );
      Markup('tbdedn','<^*','/^K:([^:]+):(.*?)$/',
  function ($m) {
    return '<:block><'.$m[1].'>'.str_replace(':+','<br /><br />',$m[2]).'</'.$m[1].'>';
  } );
      $ChapterName = array(FmtPageName('$[Chapter]', $trailname),
                           FmtPageName('$[Appendix]',$trailname));
      foreach(array('about'=>'K', 'dedication'=>'K', 'epigraph'=>'E') as $k => $v)
        if (PageVar($trailname, '$:'.$k))
            $PublishList[] = "markup:$v:p class='frontmatter':".'{$:'.$k.'}';
      Markup('tbyear','block','/^I:([^:]+):(ISBN\s+)?(.*?)\s*(?:"([^"]+)")?\s*$/',
  function ($m) {
    return '<:block><'.$m[1].'>'.$m[2].$m[3].($m[4] ? '<br />'.$m[2].$m[4] : '').'</'.$m[1].'>';
  } );
  }
  $FM = FrontMatter('default');
  foreach ($PdfFrontMatterPattern as $k => $v) SDV($GLOBALS[$v], $FM[$k]);
  $aphorisms = ExtractChapterAphorisms($trailname,$trailpage['text']);
  $dpl = $ExtractTrailStopsFunction($trailname,$trailpage['text']);
  foreach ($dpl as $k => $v) $$k = $v;

    $counter = array(); $predepth= 0; $appendix = false; $part = false;
    $chapnum = 0; $Alphabet = 'ABCDEFGHIJLKMNOPQRSTUVWXYZ';
    for($i=0;$i<count($pages);$i++) if ($depth[$i]) {
      @$counter[$depth[$i]-1]++;
      $num = $counter[0];
      for($j=$depth[$i];$j<$predepth;$j++) $counter[$j] = 0;
      $size = 0;
      for($j=1;$j<$depth[$i];$j++) { $num .= "." . $counter[$j]; $size++; }
      if ($size>3) $size=3;
      
      $anchor = "wikipublisher.$num";
      $thenum = $num;
      $num = "[[#$anchor]]";

      if ($format=='pdf') {
          $olink = $links[$i];
          $tbtag = '<'.$Tbook[$depth[$i]-1];
          if ($depth[$i]==1) {
            $linktxt = TextOfLink($trailname, $links[$i]);
            $savetxt = $linktxt;
            if (preg_match('/^[Aa]ppendix[:\s]\s*(.*)$/',$linktxt,$m)) {
              if (preg_match('/^([Uu]n)?[Pp]art:([^:]*):\s*(.*)$/',$partname.$m[1],$p)) {
                  $m[1] = $p[3];
                  $pun  = $p[1] ? $Tu : $Tp;
                  $tbtag= "<$pun>".MarkupToHTML($pagename,"H:$p[2]").$tbtag;
                  $tbtag= ($part ? "</$Tp>" : '').($appendix ? '' : "<$Ta>").$tbtag;
                  $part = true;
              } else{
                  $tbtag= ($appendix ? '' : ($part ? "</$Tp>" : '')."<$Ta>").$tbtag;
                  $part = ($appendix ? $part : false);
              }
              $olink    = str_replace($m[0],$m[1],$links[$i]);
              $savetxt  = $m[1];
              $appendix = true;
            } elseif (preg_match('/^([Aa]fterword)(:\s*(.*))?$/',$linktxt,$m)) {
              if ($m[2]) { $olink = str_replace($m[0],$m[3],$links[$i]); $savetxt  = $m[3]; }
              $linktxt  = $m[1];
              if ($part) { $part = false;  $tbtag = "</$Tp>" . $tbtag; }
            } elseif (preg_match('/^([Uu]n)?[Pp]art:([^:]*):\s*(.*)$/',MarkupEscape($partname.$linktxt),$m)) {
              $m = MarkupRestore($m);
              $olink    = str_replace($m[0],$m[3],$links[$i]);
              $savetxt  = $m[3];
              $pun      = $m[1] ? $Tu : $Tp;
              $tbtag    = ($part ? "</$Tp>" : '')."<$pun>".MarkupToHTML($pagename,"H:$m[2]").$tbtag;
              $part     = true;
            } elseif (preg_match('/^((?:[Uu]n)?[Nn]umber|[Pp](?:reface|art)):\s*(.*)$/',$linktxt,$m)) {
              $olink    = str_replace($m[0],$m[2],$links[$i]);
              $linktxt  = $m[1];
              $savetxt  = $m[2];
              if ($part) { $part = false;  $tbtag = "</$Tp>" . $tbtag; }
            }
            $partname = '';
            $linktxt = preg_replace('/\s+\(.*?\)$/','',$linktxt);
            $tbtag .= ChapterKind(strtolower($linktxt), $Unnumber);
          }
          if (preg_match('/"([^"]+)"/',$olink,$hdg))
              $olink = str_replace($hdg[0],'',$olink).$hdg[0];
          $tbtag .= '>';
          for($j=$depth[$i];$j<=$predepth;$j++) 
            $tbtag = '</'.$Tbook[$j-1].'>'.$tbtag;
          $PublishList[] = $tbtag;
          $PublishList[] = "function:PrintThisMarkup H:$num$olink";
          $contributor = trim(PageVar($pagename, '$:Contributor', $pages[$i]));
          if ($contributor) $PublishList[] = 'markup:I:author:'.$contributor;
      } else {
          if ($depth[$i]==1) {
            $linktxt = preg_replace('/^([Aa]ppendix[:\s])?/', "\\1$partname", MarkupEscape(TextOfLink($trailname,$links[$i])));
            if (preg_match('/^([Aa]ppendix[:\s])?([Uu]n)?[Pp]art:([^:]*):\s*(.*)$/',$linktxt,$m)) {
               $m = MarkupRestore($m);
               $PublishList[] = "markup:!%block class='$Tp'%".$m[3];
               $partname = '';
               if ($m[1]) {
                   $links[$i] = str_replace($m[1],'',$links[$i]);
                   $appendix = true;
                   $chapnum = 0;
              }
            } elseif (preg_match('/^([Aa]fterword|[Pp]reface|(?:[Uu]n)?[Nn]umber):\s*(.*)$/', $linktxt, $m)) {
               $links[$i] = str_replace($m[0],$m[2],$links[$i]);
               $linktxt = $m[1];
            }
            $Number = true;
            $linktxt = preg_replace('/\s+\(.*?\)$/','',$linktxt);
            if ($appendix || (ChapterKind(strtolower($linktxt),$Unnumber)==''))
                $chapnum++;
            else
                $Number = false;
            $Prefix = $ChapterName[$appendix].'&nbsp;';
            $Suffix = ':&ensp; ';
          } else {
            $Prefix = '';
            $Suffix = '&ensp; ';
          }
          $s = str_repeat('!',$size);
          $h = "!$s%block class='$Tbook[$size]'%";
          $thenum = $Number ? $Prefix.preg_replace('/^\d+/',
                                           ($appendix ? $Alphabet[$chapnum-1] : $chapnum),
                                           $thenum).$Suffix : '';
          $PublishList[] = "function:PrintThisMarkup $h$num$thenum$links[$i]";
      }      
      $predepth = $depth[$i];

      if ($pages[$i]!="textonly") {
          $PageAnchors[$pages[$i]]  = $anchor;
          $HeadingDepth[$pages[$i]] = $Tbook[$depth[$i]];
          if (isset($aphorisms[$pages[$i]])) 
              $PublishList[] = "markup:$Chapter".$aphorisms[$pages[$i]];
          if (PageExists($pages[$i])) 
	       	  $PublishList[] = "function:$PrintThisPageFunction $pages[$i]";
          elseif ($depth[$i]>1) {
              $replname = TextOfLink($trailname,$links[$i]);
              $PublishList[] = "function:PrintThisText " . 
    ((MakeLink($trailname,"[[$replname]]",NULL,NULL,'$FullName')==$pages[$i]) ?
                $links[$i] : "[[".$pages[$i]."]] ($replname)");
          }
      }
    } else $partname = $links[$i];

    if ($format=='pdf') {
        $tbtag = '';
        for($j=1;$j<=$predepth;$j++) 
            $tbtag = '</'.$Tbook[$j-1].'>'.$tbtag;
        if ($part)     $tbtag.= "</$Tp>";
        if ($appendix) $tbtag.= "</$Ta>";
        $PublishList[]  = $tbtag;
        $PublishList[0] = 
          str_replace('<chapter','<chapter id="wikipublisher.top"',$PublishList[0]);
    } else array_unshift($PublishList,'markup:[[#wikipublisher.top]]');

  $GLOBALS['PublishList'] = $PublishList;

  if ($format == 'epub')
    EbookConvert($trailname,$HandlePublishFmt);
  else
    PrintFmt($pagename,$HandlePublishFmt);
}

function ExtractChapterAphorisms($trailname,$trailtext) {
  global $SuffixPattern; $aphorisms = array();
  foreach(explode("\n",$trailtext) as $x) {
      if (preg_match("/^:\\s*([[:upper:]\\[].*)/",$x,$match)) {
          $match[1] = ReformatTrailStop(preg_replace_callback("/\\{(\\$.*?)\\}/",
                function ($m) use ($trailname) {
                  return FmtPageName($m[1], $trailname);
                },
                $match[1]));
          if (preg_match("/^(\\[\\[([^|:]*?)(?:\\|([^\\[\\]]*?(?:\\[.*?\\].*?)?))?\\]\\]($SuffixPattern))\\s*:(.*)/",
          $match[1],$m)) {
              $aphorisms[MakePageName($trailname,$m[2])] = trim($m[5]);
          }
      }
  }
  return $aphorisms;
}

function ChapterKind($kind, $att) {
  global $ChapterKindFmt;
  for ($j=0;$j<count($att);$j++)
    if ($kind==(isset($ChapterKindFmt[$att[$j]]) ? strtolower($ChapterKindFmt[$att[$j]]) : $att[$j]))
        return " kind='$att[$j]'";
  return '';
}

function SetBookBlurb($opt) {
  global  $NamePattern;
  if (preg_match("/^$NamePattern/",$opt,$m)) XLSDV('en', array('PublisherBlurb' => $m[0]));
  XLSDV('en', array('PraiseFor' => (preg_match('/"([^"]+)"$/',$opt,$m)) ? $m[1] : 'Praise for'));
  SetTmplDisplay('PageBlurbFmt', 1);          
}

function PrintMetaMarkup($pagename, $text) {
  global $PreventPageXRefs;
  $PreventPageXRefs = false;
  echo MarkupToHTML(FmtPageName('$Group.$Name',$pagename),$text);
  $PreventPageXRefs = true;
}

function EbookConvert($pagename,$HandlePublishFmt, $cover=true) {
  global $EpubDir,$EpubCacheSuffix,$UploadFileFmt,$UploadUrlFmt,$UploadPrefixFmt,
         $EpubCSSFile,$ebook_convert_path,$format;
  SDV( $EpubDir, FmtPageName($UploadFileFmt, $pagename) );
  SDV( $EpubCacheSuffix, 'html' );
  SDV( $EpubCSSFile, 'epub.css' );
  $epubfile = "$EpubDir/$pagename.$EpubCacheSuffix";
  $epubbook = "$EpubDir/$pagename.$format";
  $tmpdir   = 'tmp';
  $epubtmp  = "$EpubDir/$tmpdir";
  StopWatch( "HandlePublish(e)Book: generating $pagename" );
  ob_start();
  PrintFmt($pagename,$HandlePublishFmt);
  $EpubCachePage = ob_get_contents();
  ob_end_clean();
  mkdirp(dirname($epubfile));
  mkdirp($epubtmp);
  if ( $ep = fopen( "$epubfile,new", 'x' ) ) {
      StopWatch( "HandlePublish(e)Book: caching $pagename" );
      $EpubCachePage = EbookGetImages($EpubCachePage, $EpubDir, $tmpdir);
      fwrite( $ep, $EpubCachePage );
      fclose($ep);
      rename( "$epubfile,new", $epubfile );
      if (file_exists($epubbook)) unlink($epubbook);
      SDV( $ebook_convert_path, "/usr/bin/ebook-convert" );

      $EpubOpts[] = EbookGetCSS($EpubCachePage, $EpubDir, $tmpdir, $EpubCSSFile);
      $EpubOpts[] = '--toc-threshold "5"';
      $EpubOpts[] = "--chapter \"//*[((name()='h1' or name()='h2') and re:test(., '(chapter|book|section|part|prologue|epilogue|&sect;\d+)\s+', 'i')) or @class = 'chapter' or @class = 'section']\"";
      if (!$cover) $EpubOpts[] = '--no-default-epub-cover';

      if ( !exec("$ebook_convert_path $epubfile $epubbook ".implode(' ',$EpubOpts)) )
        Abort("HandlePublish(e)Book: error converting $epubfile > stopping");
      if ( !unlink($epubfile) )
        Abort("HandlePublish(e)Book: error unlinking $epubfile > stopping");
      StopWatch( "HandlePublish(e)Book: successfully converted $pagename" );
  } else
      Abort( "HandlePublish(e)Book: error writing cache to $epubfile,new" );
  if ( $fd = opendir($epubtmp) ) {
      while ( ( $fh = readdir($fd) ) !== FALSE ) {
        if ( $fh[0] == '.' ) continue;
        if ( !unlink("$epubtmp/$fh") )
          Abort("HandlePublish(e)Book: error unlinking $tmpdir/$fh > stopping");
      }
      if ( !rmdir($epubtmp) )
        Abort("HandlePublish(e)Book: error removing $epubtmp > stopping");
  } else
      Abort("HandlePublish(e)Book: error opening directory $epubtmp");
  Redirect($pagename);
}

function EbookGetImages($cache, $epubdir, $tmpdir) {
  return preg_replace_callback('/(<img[^>]+src=["\'])([^"\']+)(["\'][^>]*>)/',
                               function ($m) use ($epubdir, $tmpdir) {
                                 static $idx;  $idx++;
                                 $ext = preg_replace('/.*(\.[^.]+)$/',"$1",$m[2]);
                                 $fn = "$tmpdir/img$idx$ext";
                                 return copy($m[2], "$epubdir/$fn") ? 
                                                    $m[1].$fn.$m[3] :
                                                    "[error copying image]";
                               },
                               $cache);
}

function EbookGetCSS($cache, $epubdir, $tmpdir, $cssfile) {
  $fn = "$epubdir/$tmpdir/$cssfile";
  if (preg_match("|<link[^>]+href=[\"'](.*?/$cssfile)[\"'][^>]*>|",$cache,$m))
      return copy($m[1], $fn) ? "--extra-css $fn" : '';
  else
      return '--remove-paragraph-spacing';
}

?>
